unit CoreExports;

interface

uses
  Windows, SysUtils, Classes,
  TheseUtils, Temperature, TypeK;

type
  TPScanVCallback = procedure (pData: PDouble; DataSize: LongWord); stdcall;

function TcC_Convert(pRawData: PWord; RawDataCount: LongWord; pResult: PDouble; ResultCount: LongWord; ExtraGain: Double; pConfig: PByte; ConfigBytes: LongWord): LongWord; stdcall;
function TcC_StopAcquisition: LongWord; stdcall;
function TcC_StartAcquisition(TcRange, CJCRange, Oversample: Byte; AcqHz: Double; BufferSize, BaseBufferCount: LongWord; pCallback: TPScanVCallback): LongWord; stdcall;
function TcC_GetCount: Longword; stdcall;



implementation

uses
  Math,
  AIOUSB,
  AbortThreads;

const
  NAN = 0/0;
  LM335_CFrommV = 1/10;

type
  TRangeControl = record
    BiOff: LongWord;
    M: TReal;
  end;
  TPRangeControl = ^TRangeControl;

  TUpperData = record
    LeftoverData: AnsiString;
    Oversample, ScanSize: LongWord;
    TcRange, CJCRange: TRangeControl;
    pCallback: TPScanVCallback;
  end;
  TPUpperData = ^TUpperData;

const
  BaseUpperData: TUpperData = (LeftoverData: ''; Oversample: 1; ScanSize: 128*SizeOf(Word); TcRange: (BiOff: $8000; M: 1/65536); CJCRange: (BiOff: $8000; M: 1/65536); pCallback: nil);

var
  UpperData: TUpperData;
  CallbackCt: LongWord;

procedure ADBaseCallback(pBuf: PWord; BufSize, Flags, Context: LongWord); cdecl; forward;



function TcC_Convert(pRawData: PWord; RawDataCount: LongWord; pResult: PDouble; ResultCount: LongWord; ExtraGain: Double; pConfig: PByte; ConfigBytes: LongWord): LongWord;
var
  iCh, iSample: Integer;
  pThisRawData: PWord;
  pThisResult: PDouble;
  Config: array[0..19] of Byte;
  RangeCode: Byte;
  CountScale, F: TReal;
  ChmV: array of TReal;
  CJCmV: TReal;
  Oversample, SampleTot: LongWord;

begin
  Result := ERROR_SUCCESS;
  try

    if ConfigBytes < SizeOf(Config) then begin
      Result := ERROR_INSUFFICIENT_BUFFER;
      Exit;
    end;

    Move(pConfig^, Config[0], SizeOf(Config));

    Oversample := 1 + Config[$13];
    CountScale := 1/65536/Oversample;
    RawDataCount := RawDataCount div Oversample;

    if RawDataCount <= 8 then begin
      Result := ERROR_INSUFFICIENT_BUFFER;
      Exit;
    end;

    SetLength(ChmV, RawDataCount);
    pThisRawData := pRawData;
    for iCh := 0 to RawDataCount-1 do begin
      SampleTot := 0;
      for iSample := 0 to Oversample-1 do begin
        SampleTot := SampleTot + pThisRawData^;
        Inc(pThisRawData);
      end;
      F := CountScale * SampleTot;

      RangeCode := Config[iCh div 8];
      if (RangeCode and 1) <> 0 then F := F * 2 - 1;
      if (RangeCode and 2) = 0 then F := F * 2;
      if (RangeCode and 4) = 0 then F := F * 5;
      ChmV[iCh] := 1000 * F;
    end;

    CJCmV := TypeK.mVFromF(Temperature.FFromC(LM335_CFrommV * ChmV[8]));

    pThisResult := pResult;
    for iCh := 0 to Min(RawDataCount, ResultCount)-1 do begin
      F := ChmV[iCh];

      case iCh of
        8: begin
          F := LM335_CFrommV * F;
        end;
        9..15: begin
          //No change, we just report mV.
        end;
        else begin
          F := ChmV[iCh] / ExtraGain;
          F := Temperature.CFromF(TypeK.FFrommV(CJCmV + F));
        end;
      end;

      pThisResult^ := F;

      Inc(pThisResult);
    end;
    SetLength(ChmV, 0);

    for iCh := RawDataCount to ResultCount-1 do begin
      pThisResult^ := NAN;
      Inc(pThisResult);
    end;

  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;



function BuildRangeControl(RangeCode: Byte; Oversample: LongWord): TRangeControl;
begin
  Result.M := 1000/(65536*Oversample);
  if (RangeCode and 1) <> 0 then Result.BiOff := $8000 * Oversample;
  if (RangeCode and 2) = 0 then Result.M := Result.M * 2;
  if (RangeCode and 4) = 0 then Result.M := Result.M * 5;
end;

function DeviceValid: LongWord;
const
  DeviceIndex = diOnly;

var
  PID: LongWord;

begin
  try
    Result := QueryDeviceInfo(DeviceIndex, @PID, nil, nil, nil, nil);
    if Result <> ERROR_SUCCESS then Exit;

    if (PID and $FEC0) <> $8040 then Result := ERROR_BAD_TOKEN_TYPE;

  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function TcC_StopAcquisition: LongWord;
const
  DeviceIndex = diOnly;

var
  MyAbortThread: TAbortThread;

begin
  Result := DeviceValid;
  if Result = ERROR_SUCCESS then begin
    MyAbortThread := TAbortThread.BeginAbortWait(DeviceIndex);
    Result := ADC_BulkContinuousEnd(DeviceIndex, nil);
    MyAbortThread.EndAbortWaitAndFree;
    if Result = ERROR_HANDLE_EOF then Result := ERROR_SUCCESS;

    UpperData := BaseUpperData;
  end;
end;

function TcC_StartAcquisition(TcRange, CJCRange, Oversample: Byte; AcqHz: Double; BufferSize, BaseBufferCount: LongWord; pCallback: TPScanVCallback): LongWord;
const
  DeviceIndex = diOnly;

var
  Config: array of Byte;
  ConfigBufSize: LongWord;
  SpeedCode: Byte;
  I: Integer;
  Hz: Double;

begin
  try
    //Stop a previous acquisition, if any.
    Result := TcC_StopAcquisition;
    if Result <> ERROR_SUCCESS then Exit;

    //Stop the counter, in case it was running.
    Hz := 0;
    CTR_StartOutputFreq(DeviceIndex, 0, @Hz);

    //Set A/D speed.
    SpeedCode := $81;
    Result := GenericVendorWrite(DeviceIndex, $A0, 0, $401E, SizeOf(SpeedCode), @SpeedCode);
    if Result <> ERROR_SUCCESS then Exit;

    //Configure A/D.
    ConfigBufSize := 21;
    SetLength(Config, ConfigBufSize);
    for I := $0 to $F do
      Config[I] := TcRange
    ;
    Config[1] := CJCRange;
    Config[$10] := 0; //Take actual data, not internal calibration sources.
    Config[$11] := $15; //Scan selected channels each counter rising edge. (Ctr0 external.)
    Config[$12] := $F0; //Select all 128 channels.
    Config[$14] := $70; //"
    Config[$13] := Oversample;
    Result := ADC_SetConfig(DeviceIndex, @Config[0], ConfigBufSize);
    if Result <> ERROR_SUCCESS then Exit;

    //Calibrate.
    ADC_SetCal(DeviceIndex, ':NORM:');

    //Set up for callback callbacks.
    UpperData.pCallback := nil;
    UpperData.LeftoverData := '';
    UpperData.Oversample := 1+Oversample;
    UpperData.ScanSize := UpperData.Oversample * 128 * SizeOf(Word);
    UpperData.TcRange := BuildRangeControl(TcRange, UpperData.Oversample);
    UpperData.CJCRange := BuildRangeControl(CJCRange, UpperData.Oversample);
    UpperData.pCallback := pCallback;

    CallbackCt := 0;

    //Begin reading the A/D.
    Result := ADC_BulkContinuousCallbackStart(DeviceIndex, BufferSize, BaseBufferCount, 0, @ADBaseCallBack);
    if Result <> ERROR_SUCCESS then Exit;

    //Start the counter to trigger A/D scans.
    Hz := AcqHz;
    Result := CTR_StartOutputFreq(DeviceIndex, 0, @Hz);
    if Result <> ERROR_SUCCESS then Exit;



  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

procedure ADBaseCallback(pBuf: PWord; BufSize, Flags, Context: LongWord); cdecl;
const
  UnExtraGain = 1/100;
  
var
  MyUpperData: TUpperData;
  CompositeBuf: AnsiString;
  WholeScans, LeftoverLen: LongWord;
  iScan, iCh, iSample: Integer;
  F: TReal;
  FData: array of Double;
  CJCmV: TReal;
  SampleTot: LongWord;
  RangeControl: TPRangeControl;

begin
  //We don't care about Flags.
  if pBuf = nil then Exit;
  if BufSize = 0 then Exit;

  MyUpperData := UpperData;

  //Only if we've set up.
  if not Assigned(MyUpperData.pCallback) then Exit;

  LeftoverLen := Length(MyUpperData.LeftoverData);
  SetString(CompositeBuf, nil, LeftoverLen + BufSize);
  if LeftoverLen <> 0 then Move(MyUpperData.LeftoverData[1], CompositeBuf[1], LeftoverLen); MyUpperData.LeftoverData := '';
  Move(pBuf^, CompositeBuf[1 + LeftoverLen], BufSize);

  pBuf := @CompositeBuf[1];
  BufSize := LeftoverLen + BufSize;
  WholeScans := BufSize div MyUpperData.ScanSize;
  LeftoverLen := BufSize mod MyUpperData.ScanSize;
  if LeftoverLen <> 0 then
    SetString(MyUpperData.LeftoverData, @CompositeBuf[1 + WholeScans * MyUpperData.ScanSize], LeftoverLen)
  ;
  UpperData := MyUpperData;

  SetLength(FData, 128);

  for iScan := 0 to WholeScans-1 do begin

    for iCh := 0 to High(FData) do begin
      SampleTot := 0;
      for iSample := 0 to MyUpperData.Oversample-1 do begin
        SampleTot := SampleTot + pBuf^;
        Inc(pBuf);
      end;

      if (iCh div 8) = 1 then
        RangeControl := @MyUpperData.CJCRange
      else
        RangeControl := @MyUpperData.TcRange
      ;
      FData[iCh] := (SampleTot - RangeControl.BiOff) * RangeControl.M;
    end;

    FData[8] := LM335_CFrommV * FData[8];
    CJCmV := TypeK.mVFromF(Temperature.FFromC(FData[8]));

    for iCh := 0 to High(FData) do begin
      if (iCh >= 8) and (iCh <= 15) then Continue; //No change. The CJC is already converted, and for the secondary channels we just report mV.

      F := FData[iCh] * UnExtraGain;
      FData[iCh] := Temperature.CFromF(TypeK.FFrommV(CJCmV + F));
    end;

    Inc(CallbackCt);
    try
      MyUpperData.pCallback(@FData[0], Length(FData) * SizeOf(FData[0]));
    except
      {
      on E: Exception do begin
        UpperData.pCallback := nil;
        AnsiMessageBox(0, 'Aborting acquisition due to error:'#13#10 + E.Message, E.ClassName, MB_ICONEXCLAMATION);
        Exit;
      end;
      }
    end;
  end;
end;

function TcC_GetCount: Longword; stdcall;
begin
  Result := CallbackCt;
end;

initialization
  UpperData := BaseUpperData;
end.
